@@ -1,8 +1,11 @@
 import asyncio
 import re
 from http import HTTPStatus
+from io import BytesIO
+from typing import Any
 from typing import Awaitable
 from typing import Callable
+from typing import Dict
 from typing import Optional
 from urllib.parse import quote
 from urllib.parse import urlencode
@@ -12,14 +15,16 @@ import aiohttp
 from aiohttp import ClientResponse
 from aiohttp import ClientSession
 from pydantic import AnyHttpUrl
+from pydantic import field_validator
 
-from clean_python import ctx
+from clean_python import Conflict
 from clean_python import Json
+from clean_python import ValueObject
 
 from .exceptions import ApiException
 from .response import Response
 
-__all__ = ["ApiProvider"]
+__all__ = ["ApiProvider", "FileFormPost"]
 
 
 RETRY_STATUSES = frozenset({413, 429, 503})  # like in urllib3
@@ -30,6 +35,13 @@ def is_success(status: HTTPStatus) -> bool:
     return (int(status) // 100) == 2
 
 
+def check_exception(status: HTTPStatus, body: Json) -> None:
+    if status == HTTPStatus.CONFLICT:
+        raise Conflict(body.get("message", str(body)))
+    elif not is_success(status):
+        raise ApiException(body, status=status)
+
+
 JSON_CONTENT_TYPE_REGEX = re.compile(r"^application\/[^+]*[+]?(json);?.*$")
 
 
@@ -39,12 +51,14 @@ def is_json_content_type(content_type: Optional[str]) -> bool:
     return bool(JSON_CONTENT_TYPE_REGEX.match(content_type))
 
 
-def join(url: str, path: str) -> str:
+def join(url: str, path: str, trailing_slash: bool = False) -> str:
     """Results in a full url without trailing slash"""
     assert url.endswith("/")
     assert not path.startswith("/")
     result = urljoin(url, path)
-    if result.endswith("/"):
+    if trailing_slash and not result.endswith("/"):
+        result = result + "/"
+    elif not trailing_slash and result.endswith("/"):
         result = result[:-1]
     return result
 
@@ -55,6 +69,21 @@ def add_query_params(url: str, params: Optional[Json]) -> str:
     return url + "?" + urlencode(params, doseq=True)
 
 
+class FileFormPost(ValueObject):
+    file_name: str
+    file: Any  # typing of BinaryIO / BytesIO is hard!
+    field_name: str = "file"
+    content_type: str = "application/octet-stream"
+
+    @field_validator("file")
+    @classmethod
+    def validate_file(cls, v):
+        if isinstance(v, bytes):
+            return BytesIO(v)
+        assert hasattr(v, "read")  # poor-mans BinaryIO validation
+        return v
+
+
 class ApiProvider:
     """Basic JSON API provider with retry policy and bearer tokens.
 
@@ -62,24 +91,28 @@ class ApiProvider:
 
     Args:
         url: The url of the API (with trailing slash)
-        fetch_token: Callable that returns a token for a tenant id
+        headers_factory: Coroutine that returns headers (for e.g. authorization)
         retries: Total number of retries per request
         backoff_factor: Multiplier for retry delay times (1, 2, 4, ...)
+        trailing_slash: Wether to automatically add or remove trailing slashes.
     """
 
     def __init__(
         self,
         url: AnyHttpUrl,
-        fetch_token: Callable[[ClientSession, int], Awaitable[Optional[str]]],
+        headers_factory: Optional[Callable[[], Awaitable[Dict[str, str]]]] = None,
         retries: int = 3,
         backoff_factor: float = 1.0,
+        trailing_slash: bool = False,
     ):
         self._url = str(url)
-        assert self._url.endswith("/")
-        self._fetch_token = fetch_token
+        if not self._url.endswith("/"):
+            self._url += "/"
+        self._headers_factory = headers_factory
         assert retries > 0
         self._retries = retries
         self._backoff_factor = backoff_factor
+        self._trailing_slash = trailing_slash
         self._session = ClientSession()
 
     async def _request_with_retry(
@@ -89,29 +122,29 @@ class ApiProvider:
         params: Optional[Json],
         json: Optional[Json],
         fields: Optional[Json],
+        file: Optional[FileFormPost],
         timeout: float,
     ) -> ClientResponse:
-        assert ctx.tenant is not None
-        headers = {}
+        if file is not None:
+            raise NotImplementedError("ApiProvider doesn't yet support file uploads")
         request_kwargs = {
             "method": method,
-            "url": add_query_params(join(self._url, quote(path)), params),
+            "url": add_query_params(
+                join(self._url, quote(path), self._trailing_slash), params
+            ),
             "timeout": timeout,
             "json": json,
             "data": fields,
         }
-        token = await self._fetch_token(self._session, ctx.tenant.id)
-        if token is not None:
-            headers["Authorization"] = f"Bearer {token}"
+        if self._headers_factory is not None:
+            request_kwargs["headers"] = await self._headers_factory()
         for attempt in range(self._retries):
             if attempt > 0:
                 backoff = self._backoff_factor * 2 ** (attempt - 1)
                 await asyncio.sleep(backoff)
 
             try:
-                response = await self._session.request(
-                    headers=headers, **request_kwargs
-                )
+                response = await self._session.request(**request_kwargs)
                 await response.read()
             except (aiohttp.ClientError, asyncio.exceptions.TimeoutError):
                 if attempt == self._retries - 1:
@@ -129,10 +162,11 @@ class ApiProvider:
         params: Optional[Json] = None,
         json: Optional[Json] = None,
         fields: Optional[Json] = None,
+        file: Optional[FileFormPost] = None,
         timeout: float = 5.0,
     ) -> Optional[Json]:
         response = await self._request_with_retry(
-            method, path, params, json, fields, timeout
+            method, path, params, json, fields, file, timeout
         )
         status = HTTPStatus(response.status)
         content_type = response.headers.get("Content-Type")
@@ -143,10 +177,8 @@ class ApiProvider:
                 f"Unexpected content type '{content_type}'", status=status
             )
         body = await response.json()
-        if is_success(status):
-            return body
-        else:
-            raise ApiException(body, status=status)
+        check_exception(status, body)
+        return body
 
     async def request_raw(
         self,
@@ -155,10 +187,11 @@ class ApiProvider:
         params: Optional[Json] = None,
         json: Optional[Json] = None,
         fields: Optional[Json] = None,
+        file: Optional[FileFormPost] = None,
         timeout: float = 5.0,
     ) -> Response:
         response = await self._request_with_retry(
-            method, path, params, json, fields, timeout
+            method, path, params, json, fields, file, timeout
         )
         return Response(
             status=response.status,
